import {
  EnvironmentComponent,
  MouseComponent,
  ViewportComponent,
  EventEmitter,
  GameRunner,
  WebGLRenderer,
  LevelComponent,
  WorldFile,
  Engine,
} from '@freeblox/engine';
import { EditorEvents } from '../types/events';
import { WorkspaceComponent } from './workspace';
import { HistoryComponent } from './history';
import { ShortcutsComponent } from './shortcuts';

export class Editor extends Engine {
  public lastTick = performance.now();
  public events = new EventEmitter<EditorEvents>();
  public running = false;

  override mount(element: HTMLElement) {
    this.element = element;
    this.render = new WebGLRenderer(element);

    this.use(HistoryComponent);
    this.use(ViewportComponent);
    this.use(EnvironmentComponent);
    this.use(LevelComponent);
    this.use(WorkspaceComponent);
    this.use(ShortcutsComponent);
    this.use(MouseComponent);

    this.getComponent(ViewportComponent).setSizeFromViewport();
    this.start();
  }

  override loop(now: DOMHighResTimeStamp) {
    const delta = this.getDelta(now);
    this.running && requestAnimationFrame((ts) => this.loop(ts));

    this.update(delta);

    this.render.render();
    this.getComponent(WorkspaceComponent).render();
  }

  override start() {
    this.running = true;
    this.loop(this.lastTick);
    this.events.emit('initialized');
  }

  /**
   * Get current scene tree for the level, starting with two virtual
   * objects (world and environment).
   */
  public getSceneTree() {
    return this.getComponent(LevelComponent).getSceneTree();
  }

  /**
   * Serialize edited world into distributable format (JSON).
   * @param name World name
   */
  public export(name: string) {
    return this.getComponent(LevelComponent).serializeLevelSave(name);
  }

  /**
   * Deserialize world file and load it into the editor
   * @param data World data
   */
  public load(data: WorldFile) {
    return this.getComponent(LevelComponent).deserializeLevelSave(data);
  }

  /**
   * Get selected objects.
   */
  public getSelection() {
    return this.getComponent(WorkspaceComponent).selection;
  }
}
