import {
  ChangeEvent,
  EngineComponent,
  EventEmitter,
  RemoveEvent,
  Renderer,
} from '@freeblox/engine';
import { EditorEvents, TransformEvent } from '../types/events';
import { Euler, Object3D, Vector3 } from 'three';

type Changes = Record<string, unknown>;
type HistoryType = [Object, Changes, number, number?];

/**
 * Manages history (undo, redo) for editor operations.
 */
export class HistoryComponent extends EngineComponent {
  public name = HistoryComponent.name;
  public cleanUpEvents?: Function;
  private history: HistoryType[] = [];
  private restory: HistoryType[] = [];

  constructor(
    protected renderer: Renderer,
    protected events: EventEmitter<EditorEvents>
  ) {
    super(renderer, events);
  }

  initialize(): void {
    this.cleanUpEvents = this.bindEvents();
  }

  update(delta: number): void {}

  dispose(): void {
    this.cleanUpEvents?.call(this);
    this.history.length = 0;
    this.restory.length = 0;
  }

  public add(object: Object, changes: Changes, union?: number) {
    this.history.push([object, changes, Date.now(), union]);
  }

  public undo() {
    const action = this.history.pop();
    if (!action) return;
    const prevState: Changes = {};

    for (const prop in action[1]) {
      const value = action[1][prop];
      const object = action[0] as any;
      if (!object) break;
      prevState[prop] = object[prop];

      if (prop === 'parent') {
        prevState[prop] = object[prop];
        const parentObject = value as Object3D;
        if (!parentObject) {
          object.removeFromParent();
          this.events.emit('sceneLeave', object);
        }
        parentObject?.attach(object);
      } else {
        if (object[prop]?.clone) prevState[prop] = object[prop].clone();
        if (object[prop].copy && prop !== 'material') {
          object[prop].copy(value);
        } else {
          object[prop] = value;
        }
      }

      this.events.emit('change', {
        applied: true,
        edited: false,
        object,
        property: prop,
        value: value,
      });
    }

    this.restory.push([action[0], prevState, Date.now(), action[3]]);

    // Multi-undo support
    if (
      !!action[3] &&
      this.history[this.history.length - 1]?.[3] === action[3]
    ) {
      this.undo();
    }
  }

  public redo() {
    const action = this.restory.pop();
    if (!action) return;
    const prevState: Changes = {};

    for (const prop in action[1]) {
      const value = action[1][prop];
      const object = action[0] as any;
      if (!object) break;
      prevState[prop] = object[prop];

      if (prop === 'parent') {
        prevState[prop] = object[prop];
        const parentObject = value as Object3D;
        if (!parentObject) {
          object.removeFromParent();
          this.events.emit('sceneLeave', object);
        }
        parentObject?.attach(object);
      } else {
        if (object[prop]?.clone) prevState[prop] = object[prop].clone();
        if (object[prop].copy && prop !== 'material') {
          object[prop].copy(value);
        } else {
          object[prop] = value;
        }
      }

      this.events.emit('change', {
        applied: true,
        edited: false,
        object,
        property: prop,
        value: value,
      });
    }

    this.history.push([action[0], prevState, Date.now(), action[3]]);

    // Multi-redo support
    if (
      !!action[3] &&
      this.restory[this.restory.length - 1]?.[3] === action[3]
    ) {
      this.redo();
    }
  }

  private bindEvents() {
    let transformed = false;
    let transforming = false;
    let transformPosition: Vector3;
    let transformRotation: Euler;
    let transformScale: Vector3;

    const transformStart = (event: TransformEvent) => {
      transformPosition = event.position.clone();
      transformRotation = event.rotation.clone();
      transformScale = event.scale.clone();
      transforming = true;
      transformed = false;
    };

    const transformEnd = (event: TransformEvent) => {
      transforming = false;
      if (transformed) {
        this.add(event.object, {
          position: transformPosition,
          rotation: transformRotation,
          scale: transformScale,
        });
      }
    };

    const changeEvent = (event: ChangeEvent) => {
      if (transforming && event.applied) {
        transformed = true;
      }
      if (event.applied) return;

      let value = (event.object as any)[event.property];
      if (value?.clone) value = value.clone();
      this.add(event.object, {
        [event.property]: value,
      });
    };

    const removeEvent = (event: RemoveEvent) => {
      if (event.applied) return;
      if (Array.isArray(event.object)) {
        const union = Date.now();
        event.object.forEach((object) =>
          this.add(object, { parent: object.parent }, union)
        );
        return;
      }

      this.add(event.object, { parent: event.object.parent });
    };

    const undo = this.undo.bind(this);
    const redo = this.redo.bind(this);

    const resetEvent = () => {
      this.history.length = 0;
      this.restory.length = 0;
    };

    const sceneJoinEvent = (event: Object3D) => {
      this.add(event, { parent: null });
    };

    this.events.addListener('change', changeEvent);
    this.events.addListener('transformStart', transformStart);
    this.events.addListener('transformEnd', transformEnd);
    this.events.addListener('reparent', removeEvent);
    this.events.addListener('remove', removeEvent);
    this.events.addListener('undo', undo);
    this.events.addListener('redo', redo);
    this.events.addListener('resetHistory', resetEvent);
    this.events.addListener('sceneJoin', sceneJoinEvent);

    return () => {
      this.events.removeEventListener('change', changeEvent);
      this.events.removeEventListener('transformStart', transformStart);
      this.events.removeEventListener('transformEnd', transformEnd);
      this.events.removeEventListener('reparent', removeEvent);
      this.events.removeEventListener('remove', removeEvent);
      this.events.removeEventListener('undo', undo);
      this.events.removeEventListener('redo', redo);
      this.events.removeEventListener('resetHistory', resetEvent);
      this.events.removeEventListener('sceneJoin', sceneJoinEvent);
    };
  }
}
