import {
  CameraControls,
  ChangeEvent,
  EngineComponent,
  Environment,
  EventEmitter,
  GameObject,
  GameObject3D,
  Group,
  MouseButtonEvent,
  Renderer,
  TransformControls,
  World,
} from '@freeblox/engine';
import {
  AxesHelper,
  Box3,
  Box3Helper,
  BoxHelper,
  Color,
  Euler,
  GridHelper,
  LineBasicMaterial,
  Material,
  MathUtils,
  Object3D,
  Vector3,
  WebGLRenderer,
} from 'three';
import {
  EditorEvents,
  SelectEvent,
  SelectionEvent,
  TransformModeEvent,
} from '../types/events';
import { ViewHelper } from 'three/addons/helpers/ViewHelper.js';

/**
 * This component does most of the work related to editing the level.
 * Acts as a middle man for most events.
 *
 * Most importantly, handles selection, the clipboard and the editor scene.
 *
 * LevelComponent must be initialized before this component.
 */
export class WorkspaceComponent extends EngineComponent {
  public name = WorkspaceComponent.name;
  public background = new Object3D();
  public world = new World();
  public helpers = new Object3D();
  private box3 = new Box3();

  public transformControls!: TransformControls;
  private grid!: GridHelper;
  private boxes: BoxHelper[] = [];
  private box3Helper!: Box3Helper;
  private axes!: AxesHelper;
  private viewHelper!: ViewHelper;

  public transformPosition?: Vector3;
  public transformRotation?: Euler;
  public transformScale?: Vector3;
  public transforming = false;

  public selection: Object3D[] = [];
  public cutOperation = false;
  public clipboard: Object3D[] = [];
  public cleanUpEvents?: Function;

  public cameraControls!: CameraControls;

  constructor(
    protected renderer: Renderer,
    protected events: EventEmitter<EditorEvents>
  ) {
    super(renderer, events);
  }

  initialize() {
    this.world = this.renderer.scene.getObjectByName('World') as World;
    this.addToScene(this.renderer.scene);
    this.initializeHelpers();
    this.initializeControls();
    this.cleanUpEvents = this.bindEvents();
  }

  update(dt: number) {
    this.boxes.forEach((box) => box.update());

    this.cameraControls?.update(dt);

    if (this.viewHelper.animating) {
      this.viewHelper.update(dt);
    }
  }

  render() {
    this.viewHelper?.render(this.renderer.renderer as WebGLRenderer);
  }

  dispose(): void {
    this.removeFromScene(this.renderer.scene);
    this.cleanUpEvents?.call(this);
    this.viewHelper?.dispose();
    this.cameraControls?.dispose();
  }

  private addToScene(scene: Object3D) {
    this.background.name = '_background';
    this.helpers.name = '_helper';
    scene.add(this.background, this.helpers);
  }

  private removeFromScene(scene: Object3D) {
    scene.remove(this.background, this.helpers);
  }

  private bindEvents() {
    let moved = false;
    let clicked = false;
    let casterDebounce: ReturnType<typeof setTimeout> | undefined;

    const mouseDownEventHandler = (ev: MouseButtonEvent) => {
      if (
        ev.position.x > this.renderer.resolution.x - 128 &&
        ev.position.y > this.renderer.resolution.y - 128
      ) {
        this.viewHelper.handleClick(ev.raw);
        return;
      }

      moved = false;
      clicked = true;
    };

    const mouseMoveEventHandler = () => {
      if (casterDebounce || !clicked) return;
      // Prevent miniscule mouse movement when clicking from preventing a selection
      casterDebounce = setTimeout(() => {
        moved = true;
      }, 100);
    };

    const mouseUpEventHandler = (event: MouseButtonEvent) => {
      clearTimeout(casterDebounce);
      casterDebounce = undefined;
      clicked = false;
      if (moved) return;

      let object = event.target?.object;
      if (object && !(object instanceof GameObject) && object.parent) {
        object = object.parent;
      }

      // Locked is for preventing mouse picking
      if ((object as GameObject3D)?.locked) return;

      this.events.emit('select', {
        object: object as GameObject,
        multi: event.control,
      });
    };

    const selectedHandler = (select: SelectionEvent) => {
      if (!select.picker) {
        if (select.multi) {
          this.selection.push(select.object);
          return;
        }
        this.selection = [select.object];
      }

      const attachTo = this.selection.find(
        (item) => !(item as GameObject).virtual
      );

      if (attachTo) {
        this.transformControls?.attach(attachTo);
      }

      this.highlightSelection();
      this.setEncapsulatingBox3();
    };

    const deselectedHandler = (select: SelectionEvent) => {
      if (!select.picker) {
        const index = this.selection.indexOf(select.object);
        this.selection.splice(index, 1);
      }

      this.highlightSelection();
      this.setEncapsulatingBox3();

      const attachTo = this.selection.find(
        (item) => !(item as GameObject).virtual
      );
      if (this.selection.length && attachTo) {
        this.transformControls?.attach(attachTo);
        return;
      }

      this.transformControls?.detach();
    };

    const selectHandler = (event: SelectEvent) => {
      if (!event.object) {
        if (!this.selection.length) return;
        const oldSelection = this.selection;
        this.selection = [];
        oldSelection.forEach((selection) =>
          this.events.emit('deselected', {
            object: selection,
            selection: [],
            picker: true,
          })
        );
        return;
      }

      let object = event.object;
      if (this.selection.includes(object)) {
        if (event.multi) {
          const index = this.selection.indexOf(object);
          this.selection.splice(index, 1);
          this.events.emit('deselected', {
            object,
            selection: this.selection,
            picker: true,
          });
          return;
        }
      }

      if (event.multi) {
        this.selection.push(object);
        this.events.emit('selected', {
          object,
          selection: this.selection,
          multi: true,
          picker: true,
        });
        return;
      }

      const notObject = this.selection.filter(
        (entry) => entry.id !== object.id
      );

      const wasEmpty = !this.selection.length;
      this.selection = [object];
      if (wasEmpty) {
        this.events.emit('selected', {
          object: object,
          selection: [object],
          picker: true,
        });
      }

      notObject.forEach((entry) =>
        this.events.emit('deselected', {
          object: entry,
          selection: this.selection,
          picker: true,
        })
      );
    };

    const transformMode = (mode: TransformModeEvent) => {
      if (!this.transformControls) return;
      if (!mode) {
        this.transformControls.enabled = false;
        return;
      }
      this.transformControls.setMode(mode);
    };

    const transformSnap = (value: number) => {
      if (!this.transformControls) return;
      this.transformControls.setTranslationSnap(value);
      this.transformControls.setScaleSnap(value);
    };

    const transformRotationSnap = (value: number) => {
      if (!this.transformControls) return;
      this.transformControls.setRotationSnap(MathUtils.degToRad(value));
    };

    const changeListener = (change: ChangeEvent) => {
      if (['position', 'rotation', 'scale'].includes(change.property)) {
        this.setEncapsulatingBox3();
      }

      if (change.object instanceof Environment) {
        this.events.emit('setEnvironment', {
          [change.property]: change.value,
        });
      }
    };

    const copyEvent = () => {
      const filteredSelection = this.selection.filter(
        (entry) => !(entry as GameObject).virtual
      );
      if (!filteredSelection.length) return;

      this.cutOperation = false;
      this.clipboard = [];
      filteredSelection.forEach((entry) => {
        this.clipboard.push(entry);
      });
    };

    const duplicateEvent = () => {
      const filteredSelection = this.selection.filter(
        (entry) => !(entry as GameObject).virtual
      );
      if (!filteredSelection.length) return;
      this.selection.length = 0;

      filteredSelection.forEach((entry) => {
        const newObject = entry.clone();
        this.world.add(newObject);
        this.selection.push(newObject);
      });

      this.selection.forEach((entry) =>
        this.events.emit('selected', {
          object: entry,
          selection: this.selection,
          multi: this.selection.length > 1,
        })
      );
    };

    const pasteEvent = (target?: Object3D) => {
      if (!this.clipboard.length) return;
      this.selection.length = 0;

      this.clipboard.forEach((entry) => {
        const newObject = this.cutOperation ? entry : entry.clone();
        (target || this.world).add(newObject);
        this.selection.push(newObject);
      });

      this.selection.forEach((entry) =>
        this.events.emit('selected', {
          object: entry,
          selection: this.selection,
          multi: this.selection.length > 1,
        })
      );

      this.cutOperation = false;
    };

    const deleteEvent = () => {
      const filteredSelection = this.selection.filter(
        (entry) => !(entry as GameObject).virtual
      );
      if (!filteredSelection.length) return;

      const toDelete = [...filteredSelection];

      filteredSelection.forEach((entry) =>
        this.events.emit('deselected', {
          object: entry,
          selection: [],
        })
      );

      this.events.emit('remove', {
        object: toDelete,
      });

      this.selection.length = 0;
    };

    const cutEvent = () => {
      const filteredSelection = this.selection.filter(
        (entry) => !(entry as GameObject).virtual
      );

      if (!filteredSelection.length) return;

      this.clipboard = [];

      filteredSelection.forEach((entry) => {
        this.clipboard.push(entry);
      });

      this.cutOperation = true;
      deleteEvent();
    };

    const undoEvent = () => {
      if (this.cutOperation) this.cutOperation = false;
    };

    const resetEvent = () => {
      const oldSelection = [...this.selection];
      this.cutOperation = false;
      this.transforming = false;
      this.selection.length = 0;
      this.clipboard.length = 0;
      this.transformPosition = undefined;
      this.transformRotation = undefined;
      this.transformScale = undefined;
      this.transforming = false;
      oldSelection.forEach((selection) =>
        this.events.emit('deselected', {
          object: selection,
          selection: [],
        })
      );
      this.events.emit('resetHistory');
    };

    const insertEvent = (item: string) => {
      this.events.once('sceneJoin', (object) => {
        this.events.emit('select', { object });
      });

      this.events.emit('instance', { type: item });
    };

    const groupEvent = () => {
      if (!this.selection.length) return;
      // Ungroup
      if (this.selection.length === 1 && this.selection[0] instanceof Group) {
        const group = this.selection[0];
        const objects = [...group.children];
        this.events.emit('reparent', {
          object: objects,
          parent: group.parent!,
        });
        this.events.emit('remove', { object: group });
        objects.forEach((obj, index) =>
          this.events.emit('select', {
            object: obj,
            multi: index !== 0,
          })
        );
        return;
      }

      // Group
      const center = new Vector3();
      const tempBox = new Box3();
      const newGroup = new Group();
      this.selection.forEach((obj) => tempBox.expandByObject(obj));
      tempBox.getCenter(center);
      newGroup.position.copy(center);
      this.world.add(newGroup);
      this.events.emit('sceneJoin', newGroup);
      this.events.emit('reparent', {
        object: this.selection,
        parent: newGroup,
      });
      this.events.emit('select', { object: newGroup });
    };

    // Deselect scene-leaves, usually by history
    const sceneLeaveEvent = (object: Object3D) => {
      if (!this.selection.length) return;
      const indexOf = this.selection.indexOf(object);
      if (indexOf > -1) {
        this.selection.splice(indexOf, 1);
        this.events.emit('deselected', {
          object,
          selection: this.selection,
        });
      }
    };

    const selectAllEvent = () => {
      if (this.selection.length) {
        const oldSelection = this.selection;
        this.selection = [];
        oldSelection.forEach((selection) =>
          this.events.emit('deselected', {
            object: selection,
            selection: [],
            picker: true,
          })
        );
      }

      const basis = this.world.children.filter(
        (obj) => obj instanceof GameObject && !obj.virtual
      );

      basis.forEach((item) => {
        this.selection.push(item);
        this.events.emit('selected', {
          object: item,
          selection: this.selection,
          multi: basis.length > 1,
          picker: true,
        });
      });
    };

    this.events.addListener('mouseDown', mouseDownEventHandler);
    this.events.addListener('mouseMove', mouseMoveEventHandler);
    this.events.addListener('mouseUp', mouseUpEventHandler);
    this.events.addListener('select', selectHandler);
    this.events.addListener('selected', selectedHandler);
    this.events.addListener('deselected', deselectedHandler);
    this.events.addListener('transformMode', transformMode);
    this.events.addListener('transformSnap', transformSnap);
    this.events.addListener('transformRotationSnap', transformRotationSnap);
    this.events.addListener('change', changeListener);
    this.events.addListener('cut', cutEvent);
    this.events.addListener('copy', copyEvent);
    this.events.addListener('paste', pasteEvent);
    this.events.addListener('delete', deleteEvent);
    this.events.addListener('undo', undoEvent);
    this.events.addListener('reset', resetEvent);
    this.events.addListener('duplicate', duplicateEvent);
    this.events.addListener('insert', insertEvent);
    this.events.addListener('group', groupEvent);
    this.events.addListener('selectAll', selectAllEvent);
    this.events.addListener('sceneLeave', sceneLeaveEvent);

    return () => {
      this.events.removeEventListener('mouseDown', mouseDownEventHandler);
      this.events.removeEventListener('mouseMove', mouseMoveEventHandler);
      this.events.removeEventListener('mouseUp', mouseUpEventHandler);
      this.events.removeEventListener('select', selectHandler);
      this.events.removeEventListener('selected', selectedHandler);
      this.events.removeEventListener('deselected', deselectedHandler);
      this.events.removeEventListener('transformMode', transformMode);
      this.events.removeEventListener('transformSnap', transformSnap);
      this.events.removeEventListener(
        'transformRotationSnap',
        transformRotationSnap
      );
      this.events.removeEventListener('change', changeListener);
      this.events.removeEventListener('cut', cutEvent);
      this.events.removeEventListener('copy', copyEvent);
      this.events.removeEventListener('paste', pasteEvent);
      this.events.removeEventListener('delete', deleteEvent);
      this.events.removeEventListener('undo', undoEvent);
      this.events.removeEventListener('reset', resetEvent);
      this.events.removeEventListener('duplicate', duplicateEvent);
      this.events.removeEventListener('insert', insertEvent);
      this.events.removeEventListener('group', groupEvent);
      this.events.removeEventListener('selectAll', selectAllEvent);
      this.events.removeEventListener('sceneLeave', sceneLeaveEvent);
    };
  }

  private initializeHelpers() {
    this.viewHelper = new ViewHelper(
      this.renderer.camera,
      this.renderer.renderer.domElement
    );

    this.grid = new GridHelper(100, 100);
    this.background.add(this.grid);

    this.axes = new AxesHelper(50);
    this.helpers.add(this.axes);
  }

  private makeSelectBox() {
    const box = new BoxHelper(this.world, 0x00a2ff);
    (box.material as Material).depthTest = false;
    (box.material as Material).transparent = true;
    (box.material as LineBasicMaterial).linewidth = 2;
    this.helpers.add(box);
    return box;
  }

  private highlightSelection() {
    this.boxes.forEach((box) => {
      box.removeFromParent();
      box.dispose();
    });
    this.boxes.length = 0;
    if (!this.selection.length) return;

    for (const selected of this.selection) {
      if ((selected as GameObject).virtual) continue;
      const box = this.makeSelectBox();
      box.setFromObject(selected);
      this.boxes.push(box);
    }
  }

  private setEncapsulatingBox3() {
    this.box3.makeEmpty();
    if (this.box3Helper) {
      this.helpers.remove(this.box3Helper);
      this.box3Helper.dispose();
    }

    const nonVirtual = this.selection.filter(
      (item) => !(item as GameObject).virtual
    );
    if (!nonVirtual.length) return;

    nonVirtual.forEach((object) => this.box3.expandByObject(object));

    if (nonVirtual.length <= 1) return;
    this.box3Helper = new Box3Helper(this.box3, new Color(0x00a2ff));
    (this.box3Helper.material as Material).depthTest = false;
    (this.box3Helper.material as Material).transparent = true;
    (this.box3Helper.material as LineBasicMaterial).linewidth = 4;
    this.helpers.add(this.box3Helper);
  }

  private initializeControls() {
    let translationSnap: number | null = 0.5;
    let scaleSnap: number | null = 0.5;
    let rotationSnap: number | null = MathUtils.degToRad(15);
    let mode: 'translate' | 'rotate' | 'scale' = 'translate';

    if (this.cameraControls) this.cameraControls.dispose();
    if (this.transformControls) {
      translationSnap = this.transformControls.translationSnap;
      scaleSnap = this.transformControls.scaleSnap;
      rotationSnap = this.transformControls.rotationSnap;
      mode = this.transformControls.getMode();
      this.helpers.remove(this.transformControls);
      this.transformControls.dispose();
    }

    this.cameraControls = new CameraControls(
      this.renderer.camera,
      this.renderer.renderer.domElement
    );
    this.transformControls = new TransformControls(
      this.renderer.camera,
      this.renderer.renderer.domElement
    );
    this.transformControls.translationSnap = translationSnap;
    (this.transformControls as any).scaleSnap = scaleSnap;
    this.transformControls.rotationSnap = rotationSnap;
    this.transformControls.setMode(mode);
    this.helpers.add(this.transformControls);

    this.transformControls.addEventListener('mouseDown', () => {
      const target = this.transformControls.object;
      if (!target) return;
      this.transforming = true;
      this.transformPosition = target.position.clone();
      this.transformRotation = target.rotation.clone();
      this.transformScale = target.scale.clone();
      this.events.emit('transformStart', {
        position: this.transformPosition!,
        rotation: this.transformRotation!,
        scale: this.transformScale!,
        object: target,
      });
    });

    this.transformControls.addEventListener('mouseUp', () => {
      this.transforming = false;
      const target = this.transformControls.object;
      if (!target) return;
      this.events.emit('transformEnd', {
        position: target.position.clone(),
        rotation: target.rotation.clone(),
        scale: target.scale.clone(),
        object: target,
      });
    });

    this.transformControls.addEventListener('change', () => {
      const target = this.transformControls.object;
      if (!target || !this.transforming) return;

      this.events.emit('transformChange', {
        lastPosition: this.transformPosition!,
        lastRotation: this.transformRotation!,
        lastScale: this.transformScale!,
        position: target.position.clone(),
        rotation: target.rotation.clone(),
        scale: target.scale.clone(),
        object: target,
      });

      if (
        this.transformPosition &&
        !this.transformPosition.equals(target.position)
      ) {
        this.events.emit('change', {
          object: target,
          property: 'position',
          value: target.position.clone(),
          applied: true,
        });
        this.transformPosition = target.position.clone();
      }

      if (
        this.transformRotation &&
        !this.transformRotation.equals(target.rotation)
      ) {
        this.events.emit('change', {
          object: target,
          property: 'rotation',
          value: target.rotation.clone(),
          applied: true,
        });
        this.transformRotation = target.rotation.clone();
      }

      if (this.transformScale && !this.transformScale.equals(target.scale)) {
        this.events.emit('change', {
          object: target,
          property: 'scale',
          value: target.scale.clone(),
          applied: true,
        });
        this.transformScale = target.scale.clone();
      }
    });
  }
}
